# -*- coding: utf-8 -*-

""" Implementations of assorted product fetchers """

import copy
import logging
from argparse import Namespace
from concurrent.futures import ThreadPoolExecutor
from pathlib import Path
from typing import Dict

# pylint: disable=C0103, E0402, R0201, R0902, R0903, W0703, W1202, W1203

from datafetcher.errors import NGASFetchError

from .file_retrievers import NGASFileRetriever

logger = logging.getLogger(__name__)


class BaseFetcher:
    """ This is a base class for fetchers. """

    def __init__(self, args: Namespace, df_capo_settings: dict, servers_report: dict):
        self.args = args
        self.output_dir = self.args.output_dir
        self.force_overwrite = args.force
        self.dry_run = args.dry_run
        self.servers_report = servers_report
        self.settings = df_capo_settings
        self.ngas_retriever = NGASFileRetriever(self.args)
        self.retrieved = []
        self.num_files_retrieved = 0

    def retrieve_files(self, server, retrieve_method, file_specs):
        """ This is the part where we actually fetch the files. """

        retriever = NGASFileRetriever(self.args)
        num_files = len(file_specs)
        count = 0

        for file_spec in file_specs:
            count += 1

            logger.debug(
                f">>> retrieving {file_spec['relative_path']} "
                f"({file_spec['size']} bytes, "
                f"no. {count} of {num_files})...."
            )
            self.retrieved.append(retriever.retrieve(server, retrieve_method, file_spec))

        return num_files


class SerialFetcher(BaseFetcher):
    """Pull the files out, one right after another;
    don't try to be clever about it.

    """

    def __init__(self, args: Namespace, df_capo_settings: Dict, servers_report: Dict):
        super().__init__(args, df_capo_settings, servers_report)

    def run(self):
        """ fetch 'em """

        logger.debug("writing to {}".format(self.output_dir))
        logger.debug("dry run: {}".format(self.dry_run))
        logger.debug(f"force overwrite: {self.force_overwrite}")
        for server in self.servers_report:
            self.retrieve_files(
                server,
                self.servers_report[server]["retrieve_method"],
                self.servers_report[server]["files"],
            )


class ParallelFetcher(BaseFetcher):
    """ Pull the files out in parallel; try to be clever about it. """

    def __init__(self, args: Namespace, df_capo_settings: dict, servers_report: dict):
        super().__init__(args, df_capo_settings, servers_report)
        self.num_files_expected = self._count_files_expected()
        self.bucketized_files = self._bucketize_files()

    def _count_files_expected(self):
        """ determine how many files we expect to retrieve """

        count = 0
        for server in self.servers_report:
            count += len(self.servers_report[server]["files"])
        return count

    def _bucketize_files(self):
        """Takes the servers report and splits it up into a list of buckets,
        each of which a single thread will handle. There will be X * Y buckets,
        where X is the number of servers and Y is the 'threads per host', and
        the files for a given server will be distributed among the buckets
        for that server.

        Basically what we are doing here is splitting up the work among the
        threads we'll be creating and creating a list of work
        for each thread to do.
        """

        threads_per_host = int(self.settings["threads_per_host"])
        result = list()
        for server in self.servers_report:
            # Setup the 'buckets', one per server.
            bucket = {
                "server": server,
                "retrieve_method": self.servers_report[server]["retrieve_method"],
                "files": list(),
            }
            buckets = [copy.deepcopy(bucket) for _ in range(threads_per_host)]
            # Spread the files for a given server around its buckets.
            i = 0
            for file_spec in self.servers_report[server]["files"]:
                list_number = i % threads_per_host
                buckets[list_number]["files"].append(file_spec)
                i += 1
            # Trim out every bucket with no files, add the rest to the result.
            result.extend([bucket for bucket in buckets if len(bucket["files"]) > 0])
        return result

    def fetch_bucket(self, bucket):
        """ Grab the files in this bucket """
        file_sizes = [file["size"] for file in bucket["files"]]
        logger.debug(
            f"{bucket['retrieve_method']} "
            f"{len(bucket['files'])} files from "
            f"{bucket['server']}...."
        )
        num_files = self.retrieve_files(
            bucket["server"], bucket["retrieve_method"], bucket["files"]
        )
        print(f"done retrieving files {file_sizes} from {bucket['server']}")
        return num_files

    def run(self):
        """ Fetch all the files for the product locator """

        if self.args.dry_run:
            logger.debug("This is a dry run; files will not be fetched")

        with ThreadPoolExecutor() as executor:
            results = executor.map(self.fetch_bucket, self.bucketized_files)
            print(f"results: {results}")
            try:
                for result in results:
                    print(f"result has arrived: {result}")
                    self.num_files_retrieved += result
                    if self.num_files_retrieved != self.num_files_expected:
                        logger.error(
                            f"{self.num_files_expected} files expected, "
                            f"but only {self.num_files_retrieved} retrieved"
                        )
            except AttributeError:
                # (This error sometimes gets thrown after all files
                # actually -have- been retrieved. I blame the NGAS API. - JLG)

                output_path = Path(self.args.output_dir)
                files = [
                    file
                    for file in output_path.rglob("*")
                    if not file.is_dir()
                    and not str(file).endswith(".json")
                    and not str(file).endswith(".log")
                ]
                if len(files) < self.num_files_expected:
                    logger.error(
                        f"{self.num_files_expected} files expected, but only "
                        f"{self.num_files_retrieved} found"
                    )
                    raise NGASFetchError
