""" Tests of datafetcher return codes:
    check that arguments passed return expected codes
"""

# pylint: disable=C0115, C0116, E0401, E0402, E1120,R0801, R0902, R0903, R0914,
# pylint: disable=W0212, W0611, W0613, W0621, W0703, W1203

import os
import tempfile
from pathlib import Path
from typing import List

import pytest

from datafetcher.errors import (
    NoProfileException,
    MissingSettingsException,
    LocationServiceTimeoutException,
    NGASServiceRedirectsException,
    NGASFetchError,
    NoLocatorException,
    SizeMismatchException,
)
from datafetcher.datafetcher import DataFetcher
from datafetcher.utilities import (
    get_arg_parser,
    ProductLocatorLookup,
    path_is_accessible,
)

# N.B. IJ doesn't recognize imported fixtures as being in use.
# don't let these imports (make_tempdir, settings) get disappeared.
from .df_pytest_arg_utils import parse_args

from .df_pytest_utils import (
    TEST_PROFILE,
    get_test_capo_settings,
    make_tempdir,
    capo_settings,
    settings,
    launch_datafetcher,
    RUN_ALL,
    confirm_retrieve_mode_copy,
    evaluate_args_and_capo,
)

from .mock_data_fetcher import MockProdDataFetcher


def test_launch_df(make_tempdir, settings):
    """
    Our "control"; should always pass

    :param make_tempdir:
    :param settings:
    :return:
    """
    args = [
        "--product-locator",
        settings.test_data["product_locator"],
        "--output-dir",
        str(make_tempdir),
        "--profile",
        TEST_PROFILE,
        "--dry-run",
    ]
    try_to_launch_df(settings.capo_settings, args)


def test_launch_df_no_args(settings):
    with pytest.raises(MissingSettingsException):
        launch_datafetcher([], settings.capo_settings)


@pytest.mark.skipif(not RUN_ALL, reason="debug")
def test_no_args_prints_usage_and_croaks():
    """
    If the datafetcher is called without any arguments,
        it should print the "usage" statement, then exit

    :return:

    """

    with pytest.raises(TypeError):
        datafetcher = DataFetcher()
        assert datafetcher.usage == DataFetcher._build_usage_message()


@pytest.mark.skipif(not RUN_ALL, reason="debug")
def test_omitted_profile_throws_no_profile_exc(make_tempdir, settings):
    """
    Be sure DF dies with appropriate error message
    if called without Capo profile

    :param make_tempdir: tempdir created on the fly
    :param settings: gives us test data
    :return:
    """

    # store existing CAPO_PROFILE
    existing_capo_profile = os.environ["CAPO_PROFILE"]

    # remove existing CAPO_PROFILE
    os.environ["CAPO_PROFILE"] = ""

    # omit --profile arg
    args = [
        "--product-locator",
        settings.test_data["product_locator"],
        "--output-dir",
        str(make_tempdir),
    ]

    try:
        with pytest.raises(NoProfileException):
            launch_datafetcher(args, settings.capo_settings)
    finally:
        # restore the existing CAPO_PROFILE
        os.environ["CAPO_PROFILE"] = existing_capo_profile


@pytest.mark.skipif(not RUN_ALL, reason="debug")
def test_omitted_capo_value_throws_missing_setting_exc(capo_settings):
    args = [
        "--product-locator",
        "we're not going to get this far",
        "--output-dir",
        "--profile",
        "docker",
    ]
    with pytest.raises(MissingSettingsException):
        try_to_launch_df(capo_settings, args)


@pytest.mark.skipif(not RUN_ALL, reason="debug")
def test_invalid_capo_profile_raises_no_profile_exc(make_tempdir, settings):
    """
    Be sure DF dies with appropriate error message
    if called without Capo profile

    :param make_tempdir: tempdir created on the fly
    :param settings: source of DB settings
    :return:

    """

    locator = ProductLocatorLookup(settings.db_settings).look_up_locator_for_ext_name(
        "13B-014.sb28862036.eb29155786.56782.5720116088"
    )
    args = [
        "--profile",
        "whatevs",
        "--product-locator",
        locator,
        "--output-dir",
        str(make_tempdir),
    ]
    with pytest.raises(MissingSettingsException):
        try_to_launch_df(settings.capo_settings, args)


def we_are_in_docker():
    """
    Are we in a Docker container? If so, SHELL env var will not be set

    :return:
    """

    for key, _ in os.environ.items():
        if key == "SHELL":
            return False

    return True


@pytest.mark.skipif(not RUN_ALL, reason="debug")
def test_inaccessible_output_dir_throws_file_not_found(settings):
    tmpdir = Path("go_away")
    assert not path_is_accessible(tmpdir)

    altered_capo = get_test_capo_settings()
    altered_capo["download_dir"] = str(tmpdir)

    args = [
        "--product-locator",
        settings.test_data["product_locator"],
        "--profile",
        TEST_PROFILE,
        "--output-dir",
        str(tmpdir),
    ]

    with pytest.raises(FileNotFoundError):
        try_to_launch_df(settings.capo_settings, args)


@pytest.mark.skipif(not RUN_ALL, reason="debug")
def test_locator_plus_file_throws_missing_setting(make_tempdir, capo_settings):
    """

    We should reject invocation with both product locator -and- location
        file. One or the other, people!

    :param make_tempdir: tempdir created on the fly
    :param capo_settings: source of Capo settings

    :return:
    """

    args = [
        "--product-locator",
        "a_locator",
        "--location-file",
        "location.json",
        "--output-dir",
        str(make_tempdir),
        "--profile",
        TEST_PROFILE,
    ]
    with pytest.raises(MissingSettingsException):
        try_to_launch_df(capo_settings, args)


class MockServiceTimeoutReturn:
    """ Simulates locator request service timeout """

    @staticmethod
    def run():
        raise LocationServiceTimeoutException()


@pytest.mark.skipif(not RUN_ALL, reason="debug")
def test_locator_service_timeout_returns_expected_code(monkeypatch, settings, make_tempdir):
    def mock_run(*args, **kwargs):
        return MockServiceTimeoutReturn.run()

    args = [
        "--product-locator",
        settings.test_data["product_locator"],
        "--output-dir",
        str(make_tempdir),
        "--profile",
        TEST_PROFILE,
    ]
    monkeypatch.setattr(DataFetcher, "run", mock_run)
    with pytest.raises(LocationServiceTimeoutException):
        launch_datafetcher(args, settings.capo_settings)


class MockTooManyServiceRedirectsReturn:
    @staticmethod
    def run():
        raise NGASServiceRedirectsException()


@pytest.mark.skipif(not RUN_ALL, reason="debug")
def test_too_many_service_redirects_returns_expected_code(monkeypatch, settings, make_tempdir):
    def mock_run(*args, **kwargs):
        return MockTooManyServiceRedirectsReturn().run()

    args = [
        "--product-locator",
        settings.test_data["product_locator"],
        "--output-dir",
        str(make_tempdir),
        "--profile",
        TEST_PROFILE,
    ]
    monkeypatch.setattr(DataFetcher, "run", mock_run)
    with pytest.raises(NGASServiceRedirectsException):
        launch_datafetcher(args, settings.capo_settings)


class MockCatastrophicServiceErrorReturn:
    @staticmethod
    def run():
        raise NGASFetchError()


@pytest.mark.skipif(not RUN_ALL, reason="debug")
def test_catastrophic_service_error_returns_expected_code(monkeypatch, settings, make_tempdir):
    def mock_run(*args, **kwargs):
        return MockCatastrophicServiceErrorReturn().run()

    args = [
        "--product-locator",
        settings.test_data["product_locator"],
        "--output-dir",
        str(make_tempdir),
        "--profile",
        TEST_PROFILE,
    ]

    monkeypatch.setattr(DataFetcher, "run", mock_run)
    with pytest.raises(NGASFetchError):
        launch_datafetcher(args, settings.capo_settings)


@pytest.mark.skipif(not RUN_ALL, reason="debug")
def test_copy_attempt_raises_ngas_fetch_error(make_tempdir, settings):
    args = [
        "--product-locator",
        settings.test_data["product_locator"],
        "--output-dir",
        str(make_tempdir),
        "--profile",
        TEST_PROFILE,
    ]

    namespace = get_arg_parser().parse_args(args)
    fetcher = MockProdDataFetcher(namespace, settings.capo_settings)

    servers_report = fetcher.servers_report
    confirm_retrieve_mode_copy(servers_report)

    # let's try just one file so we're not sitting here all day
    a_server = None
    for server in servers_report:
        entry = servers_report[server]
        servers_report = {server: entry}
        fetcher.servers_report = servers_report
        assert fetcher.servers_report[server] is not None
        a_server = server
        break
    files = fetcher.servers_report[a_server]["files"]
    fetcher.servers_report[a_server]["files"] = [files[0]]

    with pytest.raises(NGASFetchError):
        fetcher.run()


@pytest.mark.skipif(not RUN_ALL, reason="debug")
def test_product_locator_not_found_returns_expected_code(make_tempdir, settings):
    args = [
        "--product-locator",
        "not_a_locator",
        "--output-dir",
        str(make_tempdir),
        "--profile",
        TEST_PROFILE,
    ]
    with pytest.raises(NoLocatorException):
        launch_datafetcher(args, settings.capo_settings)


@pytest.mark.skipif(not RUN_ALL, reason="debug")
def test_unable_to_open_location_file_throws_file_not_found(make_tempdir, capo_settings):
    args = [
        "--location-file",
        "location.json",
        "--output-dir",
        str(make_tempdir),
        "--profile",
        TEST_PROFILE,
    ]
    with pytest.raises(FileNotFoundError):
        try_to_launch_df(capo_settings, args)


class MockNgasFetchError:
    """ Simulates a problem fetching file(s) from NGAS server """

    @staticmethod
    def run():
        raise NGASFetchError()


@pytest.mark.skipif(not RUN_ALL, reason="debug")
def test_error_fetching_file_from_ngas_returns_expected_code(monkeypatch, settings, make_tempdir):
    def mock_run(*args, **kwargs):
        return MockNgasFetchError().run()

    args = [
        "--product-locator",
        settings.test_data["product_locator"],
        "--output-dir",
        str(make_tempdir),
        "--profile",
        TEST_PROFILE,
    ]
    monkeypatch.setattr(DataFetcher, "run", mock_run)
    with pytest.raises(NGASFetchError):
        launch_datafetcher(args, settings.capo_settings)


class MockSizeMismatchError:
    """ Simulates a retrieved file of unexpected size """

    @staticmethod
    def run():
        raise SizeMismatchException()


@pytest.mark.skipif(not RUN_ALL, reason="debug")
def test_unexpected_size_returns_expected_code(monkeypatch, settings, make_tempdir):
    def mock_run(*args, **kwargs):
        return MockSizeMismatchError().run()

    args = [
        "--product-locator",
        settings.test_data["product_locator"],
        "--output-dir",
        str(make_tempdir),
        "--profile",
        TEST_PROFILE,
    ]
    monkeypatch.setattr(DataFetcher, "run", mock_run)
    with pytest.raises(SizeMismatchException):
        launch_datafetcher(args, settings.capo_settings)


def try_to_launch_df(capo_settings, args: List[str]):
    parse_args(args)
    namespace = evaluate_args_and_capo(args, capo_settings)
    fetcher = DataFetcher(namespace, capo_settings)
    return fetcher.run()
