import inspect
import json
import logging
from collections import defaultdict
from functools import partial
from threading import Thread
from typing import Any, Dict

import transaction
from kombu import Message
from messaging.messenger import MessageReceiverIF, Messenger

"""
Class Router: for creating the Router object.
The router object directs events received from the Messenger object 
to the appropriate consumer methods. 
"""

logger = logging.getLogger(__name__)


def on_message(**pattern):
    """
    Decorator for subscriber methods that allows them to specify a message pattern to match

    :param pattern: Pattern for message type subscriber requires
    :return: Wrapper function containing logic for decorator
    """

    def wrapper(func):
        if hasattr(func, "on_message_patterns"):
            func.on_message_patterns.append(pattern)
        else:
            func.on_message_patterns = [pattern]
        return func

    return wrapper


class Router(MessageReceiverIF):
    def __init__(self, service: str, messenger: Messenger = None):
        self.service = service
        self.messenger = messenger if messenger else self.create_messenger()
        self.callbacks = defaultdict(list)

    def create_messenger(self):
        """
        Create and initialize messenger and corresponding thread; start the thread

        :return: Created messenger
        """
        # print("Router.create_messenger> creating messenger and thread")
        self.messenger = Messenger(self, self.service)
        self.messenger_thread = Thread(target=self.messenger.run, name="messenger")
        self.messenger_thread.daemon = True
        self.messenger_thread.start()
        return self.messenger

    def register(self, instance: object):
        """
        Register marked callback methods of a given instance of an object with this router; if a method of this
        instance has an on_message_pattern attribute, it is added to the dictionary of callback functions

        :param instance: Instance of an object
        """
        # Partial function that wraps json.dumps and adds a couple of optional arguments that will ensure consistency in
        # dictionary string representations
        dict_to_str = partial(json.dumps, sort_keys=True, ensure_ascii=True, default=str)

        members = inspect.getmembers(
            instance, lambda meth: inspect.ismethod(meth) and hasattr(meth, "on_message_patterns")
        )
        for name, member in members:
            for pattern in member.on_message_patterns:
                # Parse the pattern dict into a JSON string
                hashable_pattern_repr = dict_to_str(pattern)
                # print("Router.register> str repr of message pattern: " + hashable_pattern_repr)
                self.callbacks[hashable_pattern_repr].append(member)
                # print(f"Router.register> callbacks: {self.callbacks}")

    def receive_message(self, message: str):
        """
        Method that gets called when a message is received; iterates through callbacks and sends the received message
        to all corresponding callbacks

        :param message: Received message
        """
        if self.service == "capability":
            logger.info(f"receiving message: {message}")
        message = json.loads(message)

        for pattern, callbacks in self.callbacks.items():
            # Convert hashable str version of pattern back to dict
            dict_pattern = json.loads(pattern)

            logger.info(
                f"\nRouter.receive_message > pattern: {dict_pattern}, message: {message['type']}, matches: {self.matches(pattern, message)}\n"
            )
            if self.matches(dict_pattern, message):
                # print(f"Router.receive_message> pattern {pattern} matches with message {message}")
                for callback in callbacks:
                    callback(**message)

        transaction.commit()

    def send_message(self, **message):
        """
        Adapter method that simply forwards a message to its messenger

        :param message: AMQP message in dict format
        """
        self.messenger.send_message(**message)

    @staticmethod
    def matches(pattern: Dict, message: Dict) -> bool:
        """
        Performs a check on a given pattern and message to determine if the message fits the pattern; a message fits
        a pattern if that message has any entries from the pattern

        :param pattern: Dictionary containing entries to be matched with the message
        :param message: Dictionary containing entries conveying information
        :return: True if the message matches the pattern; else False
        """
        if type(pattern) == type(message) and type(message) == type({}):
            return all(
                key in message and Router.matches(pattern[key], message[key])
                for key in pattern.keys()
            )
        else:
            return type(pattern) == type(message) and pattern == message
